/* knapsack.c */

/* Direttive */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "knapsack.h"
#include "alloc.h"


/* Prototipi delle procedure secondarie */

void InterpretaLineaComando (int argc, char *argv[], char *filedati);

/* Algoritmo ricorsivo per il problema dello zaino */
int AlgoritmoRicorsivoKP (int n, vint phi, vint v, int V, int **F);

/* Programmazione dinamica ricorsiva per il problema dello zaino */
void PDricorsivaKP (datiKP *pI, soluzioneKP *pS);


/* Programma principale */
int main (int argc, char *argv[])
{
  /* Parte dichiarativa */
  char filedati[ROW_LENGTH];
  datiKP I;
  soluzioneKP S;
  clock_t inizio, fine;
  double tempo;


  /* Parte esecutiva */
  InterpretaLineaComando(argc,argv,filedati);

  /* Carica i dati */
  CaricaDatiKP(filedati,&I);

  /* Crea una soluzione vuota */
  CreaSoluzioneKP(I.n,&S);

  /* Risolve il problema dello zaino */
  inizio = clock();
  PDricorsivaKP(&I,&S);
  fine = clock();
  tempo = ((double) (fine - inizio))/CLOCKS_PER_SEC;

  /* Stampa il tempo di calcolo e la soluzione */
  printf("CPU = %f\n",tempo);
  StampaSoluzioneKP(I.n,&S);

  /* Dealloca le strutture dati dinamiche */
  DistruggeDatiKP(&I);
  DistruggeSoluzioneKP(&S);

  return EXIT_SUCCESS;
}


/* Definizione delle procedure secondarie */

void InterpretaLineaComando (int argc, char *argv[], char *filedati)
{
  if (argc != 2)
  {
    fprintf(stderr,"Errore nella linea di comando!\n");
    fprintf(stderr,"%s [filedati]\n",argv[0]);
    exit(EXIT_FAILURE);
  }

  strcpy(filedati,argv[1]);
}


/* Algoritmo ricorsivo per il problema dello zaino */
int AlgoritmoRicorsivoKP (int n, vint phi, vint v, int V, int **F)
{
  int f0, f1;

  if (F[n][V] < 0)
  {
    if (n == 0)
      F[n][V] = 0;
    else if (v[n] > V)
      F[n][V] = AlgoritmoRicorsivoKP(n-1,phi,v,V,F);
    else /* (v[n] <= V) */
    {
      f0 = AlgoritmoRicorsivoKP(n-1,phi,v,V,F);
      f1 = phi[n] + AlgoritmoRicorsivoKP(n-1,phi,v,V-v[n],F);
      F[n][V] = ( (f0 >= f1) ? f0 : f1 );
    }
  }

  return F[n][V];
}



/* Programmazione dinamica ricorsiva per il problema dello zaino */
void PDricorsivaKP (datiKP *pI, soluzioneKP *pS)
{
  mint F;
  int i, v;


  /* Crea la matrice ausiliaria che conserva i sottoproblemi */
  F = int2alloc(pI->n+1,pI->V+1);
  for (i = 0; i <= pI->n; i++)
    for (v = 0; v <= pI->V; v++)
      F[i][v] = -1;

  /* Programmazione dinamica ricorsiva per il problema dello zaino */
  pS->f = AlgoritmoRicorsivoKP(pI->n,pI->phi,pI->v,pI->V,F);

  /* Dealloca la matrice ausiliaria */
  int2free(&F,pI->n+1);
}
