/* knapsack.c */

/* Direttive */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "alloc.h"
#include "knapsack.h"


/* Prototipi delle procedure secondarie */

void InterpretaLineaComando (int argc, char *argv[], char *filedati);

/* Risolve il Knapsack Problem con la programmazione dinamica bottom-up (iterativa) */
void PDiterativaKP (datiKP *pI, soluzioneKP *pS);


/* Programma principale */
int main (int argc, char *argv[])
{
  /* Parte dichiarativa */
  char filedati[ROW_LENGTH];
  datiKP I;
  soluzioneKP S;
  clock_t inizio, fine;
  double tempo;


  /* Parte esecutiva */
  InterpretaLineaComando(argc,argv,filedati);

  /* Carica i dati */
  CaricaDatiKP(filedati,&I);

  /* Crea una soluzione vuota */
  CreaSoluzioneKP(I.n,&S);

  /* Risolve il problema dello zaino */
  inizio = clock();
  PDiterativaKP(&I,&S);
  fine = clock();
  tempo = ((double) (fine - inizio))/CLOCKS_PER_SEC;

  /* Stampa il tempo di calcolo e la soluzione */
  printf("CPU = %f\n",tempo);
  StampaSoluzioneKP(I.n,&S);

  /* Dealloca le strutture dati dinamiche */
  DistruggeDatiKP(&I);
  DistruggeSoluzioneKP(&S);

  return EXIT_SUCCESS;
}


/* Definizione delle procedure secondarie */

void InterpretaLineaComando (int argc, char *argv[], char *filedati)
{
  if (argc != 2)
  {
    fprintf(stderr,"Errore nella linea di comando!\n");
    fprintf(stderr,"%s [filedati]\n",argv[0]);
    exit(EXIT_FAILURE);
  }

  strcpy(filedati,argv[1]);
}


/* Risolve il Knapsack Problem con la programmazione dinamica bottom-up (iterativa) */
void PDiterativaKP (datiKP *pI, soluzioneKP *pS)
{
  int **F;
  int i, v;
  int f0, f1;


  /* Crea la matrice ausiliaria che conserva i sottoproblemi */
  F = int2alloc(pI->n+1,pI->V+1);

  /* Programmazione dinamica iterativa per il problema dello zaino */
  for (i = 0; i <= pI->n; i++)
  {
    for (v = 0; v <= pI->V; v++)
    {
      /* Risolve il sottoproblema (i,v) */
      if (i == 0)
        F[i][v] = 0;                            /* return 0;                                */
      else if (pI->v[i] > v)
        F[i][v] = F[i-1][v];                  /* return PDricorsivaKP(n-1,phi,v,V);          */
      else /* (pI->v[i] <= v) */
      {
        f0 = F[i-1][v];                       /* return PDricorsivaKP(n-1,phi,v,V);          */
        f1 = pI->phi[i] + F[i-1][v-pI->v[i]]; /* return PDricorsivaKP(n-1,phi,v,V-pI->v[n]); */
        F[i][v] = ( (f0 >= f1) ? f0 : f1 );
      }
    }
  }

  /* Il problema originale corrisponde alla cella (n,V): il suo valore ottimo e' il contenuto della cella */
  pS->f = F[pI->n][pI->V];

  /* Ricostruisce la soluzione, determinando all'indietro quale sottoproblema
     ha determinato la soluzione di ogni problema */
  v = pI->V;
  for (i = pI->n; i >= 1; i--)
    if (F[i][v] == F[i-1][v])
      pS->x[i] = FALSE;
    else
    {
      pS->x[i] = TRUE;
      v -= pI->v[i];
    }

  /* Dealloca la matrice ausiliaria */
  int2free(&F,pI->n+1);
}
